# -*- coding: utf-8 -*-
"""
babel handlers params module.
"""

import os

import pyrin.configuration.services as config_services
import pyrin.application.services as application_services

from pyrin.globalization.locale.babel.interface import BabelCLIParamBase
from pyrin.globalization.locale.babel import BabelPackage
from pyrin.cli.arguments import KeywordArgument, BooleanArgument, \
    PositionalArgument, CompositeKeywordArgument


class DomainParam(KeywordArgument, BabelCLIParamBase):
    """
    domain param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of DomainParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            defaults to `domain` value form babel
                                            config store if not provided.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'domain')

        super().__init__('domain', '--domain', default=default)


class CompileDomainsParam(CompositeKeywordArgument, BabelCLIParamBase):
    """
    compile domains param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of CompileDomainsParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            defaults to `compile_domains` value form
                                            babel config store if not provided.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'compile_domains')

        super().__init__('domain', '--domain', default=default)


class InputTemplateFileParam(KeywordArgument, BabelCLIParamBase):
    """
    input template file param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of InputTemplateFileParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            if not provided, defaults to
                                            `locale_path/template_file`
                                            where `template_file` comes from
                                            babel config store.
        """

        if default is None:
            default = os.path.join(application_services.get_locale_path(),
                                   config_services.get('babel', 'arguments', 'template_file'))

        super().__init__('input_file', '--input-file', default=default)


class OutputDirectoryParam(KeywordArgument, BabelCLIParamBase):
    """
    output directory param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of OutputDirectoryParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            if not provided, defaults to
                                            `locale_path` of application.
        """

        if default is None:
            default = application_services.get_locale_path()

        super().__init__('output_dir', '--output-dir', default=default)


class OutputTemplateFileParam(KeywordArgument, BabelCLIParamBase):
    """
    output template file param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of OutputTemplateFileParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            if not provided, defaults to
                                            `locale_path/template_file`
                                            where `template_file` comes from
                                            babel config store.
        """

        if default is None:
            default = os.path.join(application_services.get_locale_path(),
                                   config_services.get('babel', 'arguments', 'template_file'))

        super().__init__('output_file', '--output-file', default=default)


class OutputFileParam(KeywordArgument, BabelCLIParamBase):
    """
    output file param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of OutputFileParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            if not provided, defaults to None.
        """

        super().__init__('output_file', '--output-file', default=default)


class OmitHeaderParam(BooleanArgument, BabelCLIParamBase):
    """
    omit header param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of OmitHeaderParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            if not provided, defaults to `omit_header`
                                            value from babel config store.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'omit_header')

        super().__init__('omit_header', '--omit-header', default=default)


class LocaleParam(KeywordArgument, BabelCLIParamBase):
    """
    locale param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of LocaleParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            if not provided, defaults to None.
        """

        super().__init__('locale', '--locale', default=default)


class WidthParam(KeywordArgument, BabelCLIParamBase):
    """
    width param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of WidthParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            if not provided, defaults to `width`
                                            value from babel config store.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'width')

        super().__init__('width', '--width', default=default)


class NoWrapParam(BooleanArgument, BabelCLIParamBase):
    """
    no wrap param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of NoWrapParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            if not provided, defaults to `no_wrap`
                                            value from babel config store.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'no_wrap')

        super().__init__('no_wrap', '--no-wrap', default=default)


class IgnoreObsoleteParam(BooleanArgument, BabelCLIParamBase):
    """
    ignore obsolete param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of IgnoreObsoleteParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            if not provided, defaults to `ignore_obsolete`
                                            value from babel config store.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'ignore_obsolete')

        super().__init__('ignore_obsolete', '--ignore-obsolete', default=default)


class NoFuzzyMatchingParam(BooleanArgument, BabelCLIParamBase):
    """
    no fuzzy matching param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of NoFuzzyMatchingParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            if not provided, defaults to
                                            `no_fuzzy_matching`
                                            value from babel config store.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'no_fuzzy_matching')

        super().__init__('no_fuzzy_matching', '--no-fuzzy-matching', default=default)


class UpdateHeaderCommentParam(BooleanArgument, BabelCLIParamBase):
    """
    update header comment param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of UpdateHeaderCommentParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            if not provided, defaults to
                                            `update_header_comment`
                                            value from babel config store.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'update_header_comment')

        super().__init__('update_header_comment', '--update-header-comment', default=default)


class PreviousParam(BooleanArgument, BabelCLIParamBase):
    """
    previous param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of PreviousParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            if not provided, defaults to `previous`
                                            value from babel config store.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'previous')

        super().__init__('previous', '--previous', default=default)


class CharsetParam(KeywordArgument, BabelCLIParamBase):
    """
    charset param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of CharsetParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            if not provided, defaults to `charset`
                                            value from babel config store.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'charset')

        super().__init__('charset', '--charset', default=default)


class KeywordsParam(CompositeKeywordArgument, BabelCLIParamBase):
    """
    keywords param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of KeywordsParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            if not provided, defaults to `keywords`
                                            value from babel config store.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'keywords')

        super().__init__('keywords', '--keywords', default=default)


class NoDefaultKeywordsParam(BooleanArgument, BabelCLIParamBase):
    """
    no default keywords param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of NoDefaultKeywordsParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            if not provided, defaults to
                                            `no_default_keywords`
                                            value from babel config store.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'no_default_keywords')

        super().__init__('no_default_keywords', '--no-default-keywords', default=default)


class MappingParam(KeywordArgument, BabelCLIParamBase):
    """
    mapping param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of MappingParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            if not provided, defaults to
                                            path to `babel.mappings.config`.
        """

        if default is None:
            default = config_services.get_file_path(BabelPackage.EXTRA_CONFIG_STORE_NAMES[0])

        super().__init__('mapping', '--mapping', default=default)


class NoLocationParam(BooleanArgument, BabelCLIParamBase):
    """
    no location param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of NoLocationParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            if not provided, defaults to `no_location`
                                            value from babel config store.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'no_location')

        super().__init__('no_location', '--no-location', default=default)


class AddLocationParam(KeywordArgument, BabelCLIParamBase):
    """
    add location param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of AddLocationParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            if not provided, defaults to `add_location`
                                            value from babel config store.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'add_location')

        super().__init__('add_location', '--add-location', default=default)


class SortOutputParam(BooleanArgument, BabelCLIParamBase):
    """
    sort output param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of SortOutputParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            if not provided, defaults to `sort_output`
                                            value from babel config store.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'sort_output')

        super().__init__('sort_output', '--sort-output', default=default)


class SortByFileParam(BooleanArgument, BabelCLIParamBase):
    """
    sort by file param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of SortByFileParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            if not provided, defaults to `sort_by_file`
                                            value from babel config store.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'sort_by_file')

        super().__init__('sort_by_file', '--sort-by-file', default=default)


class MSGIDBugsAddressParam(KeywordArgument, BabelCLIParamBase):
    """
    msgid bugs address param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of MSGIDBugsAddressParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            if not provided, defaults to
                                            `msgid_bugs_address`
                                            value from babel config store.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'msgid_bugs_address')

        super().__init__('msgid_bugs_address', '--msgid-bugs-address', default=default)


class CopyrightHolderParam(KeywordArgument, BabelCLIParamBase):
    """
    copyright holder param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of CopyrightHolderParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            if not provided, defaults to
                                            `copyright_holder`
                                            value from babel config store.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'copyright_holder')

        super().__init__('copyright_holder', '--copyright-holder', default=default)


class ProjectParam(KeywordArgument, BabelCLIParamBase):
    """
    project param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of ProjectParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            if not provided, defaults to `project`
                                            value from babel config store.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'project')

        super().__init__('project', '--project', default=default)


class VersionParam(KeywordArgument, BabelCLIParamBase):
    """
    version param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of VersionParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            if not provided, defaults to `version`
                                            value from babel config store.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'version')

        super().__init__('version', '--version', default=default)


class AddCommentsParam(CompositeKeywordArgument, BabelCLIParamBase):
    """
    add comments param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of AddCommentsParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            if not provided, defaults to `add_comments`
                                            value from babel config store.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'add_comments')

        super().__init__('add_comments', '--add-comments', default=default)


class StripCommentsParam(BooleanArgument, BabelCLIParamBase):
    """
    strip comments param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of StripCommentsParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            if not provided, defaults to `strip_comments`
                                            value from babel config store.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'strip_comments')

        super().__init__('strip_comments', '--strip-comments', default=default)


class InputPathsParam(PositionalArgument, BabelCLIParamBase):
    """
    input paths param class.
    """

    def __init__(self, index=None, default=None):
        """
        initializes an instance of InputPathsParam.

        :param int index: zero based index of this param in cli command inputs.
                          defaults to 0 if not provided.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            defaults to both application and pyrin
                                            main package paths if not provided.
        """

        if index is None:
            index = 0

        if default is None:
            default = self._get_paths(include_application=True, include_pyrin=True)

        super().__init__('input_paths', index, default=default)

    def _process_inputs(self, **options):
        """
        processes the inputs given to this handler and returns them.

        :rtype: dict
        """

        include_application = options.get('include_app', False)
        include_pyrin = options.get('include_pyrin', False)
        paths = self._get_paths(include_application, include_pyrin)

        if len(paths) <= 0:
            paths = None

        options.update(input_paths=paths)
        return super()._process_inputs(**options)

    def _get_paths(self, include_application=True, include_pyrin=True):
        """
        gets application and pyrin main package paths based on given inputs.

        :param bool include_application: include application main package path.
                                         defaults to True if not provided.

        :param bool include_pyrin: include pyrin main package path.
                                   defaults to True if not provided.

        :returns: list[str]
        :rtype: list
        """

        paths = []
        if include_application is True:
            paths.append(application_services.get_application_main_package_path())

        if include_pyrin is True:
            paths.append(application_services.get_pyrin_main_package_path())

        return paths


class DirectoryParam(KeywordArgument, BabelCLIParamBase):
    """
    directory param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of DirectoryParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            if not provided, defaults to
                                            `locale_path` of application.
        """

        if default is None:
            default = application_services.get_locale_path()

        super().__init__('directory', '--directory', default=default)


class UseFuzzyParam(BooleanArgument, BabelCLIParamBase):
    """
    use fuzzy param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of UseFuzzyParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            if not provided, defaults to `use_fuzzy`
                                            value from babel config store.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'use_fuzzy')

        super().__init__('use_fuzzy', '--use-fuzzy', default=default)


class StatisticsParam(BooleanArgument, BabelCLIParamBase):
    """
    statistics param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of StatisticsParam.

        :param Union[object, None] default: default value to be emitted to
                                            cli if this param is not available.
                                            if set to None, this param will not
                                            be emitted at all.
                                            if not provided, defaults to None.
        """

        super().__init__('statistics', '--statistics', default=default)



# -*- coding: utf-8 -*-
"""
database orm session base module.
"""

from sqlalchemy.orm import Session
from sqlalchemy.sql.elements import TextClause

import pyrin.database.services as database_services
import pyrin.database.orm.sql.extractor.services as extractor_services

from pyrin.database.exceptions import InvalidDatabaseBindError


class CoreSession(Session):
    """
    core session class.
    all application sessions must be an instance or a subclass of this.
    """

    def execute(self, clause, params=None, mapper=None, bind=None, **kw):
        """
        executes a sql expression construct or string statement within
        the current transaction.

        the current transaction will be resolved using the given
        entity class or object or bind name.

        this method is a wrapper for `sqlalchemy.orm.session.execute()` method
        which wraps it to bind to correct engine on different situations.
        for full documentation see the `sqlalchemy.orm.session.execute()` docs.

        :param Union[Executable, str] clause: an executable statement `Executable`
                                              expression such as `expression.select`
                                              or string sql statement to be executed.

        :param Union[dict, list[dict]] params: optional dictionary, or list of
                                               dictionaries, containing bound
                                               parameter values. if a single dictionary,
                                               single-row execution occurs. if a list of
                                               dictionaries, an `executemany' will be invoked.
                                               the keys in each dictionary must correspond
                                               to parameter names present in the statement.

        :param CoreEntity mapper: entity class or object which the given clause
                                  must be executed on its related engine.
                                  this argument takes precedence over
                                  `clause` when locating a bind.

        :param Engine bind: optional engine to be used as the bind. if
                            this engine is already involved in an ongoing transaction,
                            that connection will be used. this argument takes
                            precedence over `mapper` and `clause` and `bind_name` when
                            locating a bind.

        :keyword kw: additional keyword arguments are sent to `session.get_bind()`
                     to allow extensibility of `bind` schemes.

        :keyword str bind_name: database bind name from `database.binds.config`
                                file which given clause must be executed on.
                                this keyword argument precedence over `mapper` and
                                `clause` when locating a bind.

        :raises InvalidDatabaseBindError: invalid database bind error.

        :returns: results of the statement execution.
        :rtype: ResultProxy
        """

        bind_name = kw.pop('bind_name', None)
        if bind is None and bind_name is not None:
            bind = database_services.get_bounded_engines().get(bind_name, None)
            if bind is None:
                raise InvalidDatabaseBindError('Database bind name [{bind_name}] '
                                               'is not available in '
                                               '[database.binds.config] file.'
                                               .format(bind_name=bind_name))

        return super().execute(clause, params, mapper, bind, **kw)

    def get_bind(self, mapper=None, clause=None, **kw):
        """
        returns a `bind` to which this `CoreSession` is bound.
        this method extends the `get_bind() method of `Session` by
        trying to find bounded engine to tables of a raw string query.

        the `bind` is usually an instance of `Engine`,
        except in the case where the `CoreSession` has been
        explicitly bound directly to a `Connection`.

        for a multiply-bound or unbound `CoreSession`, the
        `mapper` or `clause` arguments are used to determine the
        appropriate bind to return.

        note that the 'mapper' argument is usually present
        when `CoreSession.get_bind` is called via an orm
        operation such as a `Session.query`, each
        individual INSERT/UPDATE/DELETE operation within a
        `Session.flush`, call, etc.

        The order of resolution is:

        1. if mapper given and session.binds is present,
           locate a bind based first on the mapper in use, then
           on the mapped class in use, then on any base classes that are
           present in the `__mro__` of the mapped class, from more specific
           superclasses to more general.
        2. if clause given and session.binds is present,
           locate a bind based on `Table` objects
           found in the given clause present in session.binds.
        3. if session.bind is present, return that.
        4. if clause given, attempt to return a bind
           linked to the `MetaData` ultimately
           associated with the clause.
        5. if mapper given, attempt to return a bind
           linked to the `MetaData` ultimately
           associated with the `Table` or other
           selectable to which the mapper is mapped.
        6. no bind can be found, `sqlalchemy.exc.UnboundExecutionError`
           is raised.

        :param CoreEntity mapper: a CoreEntity instance or class. the bind can be derived from
                                  a `Mapper` first by consulting the `binds` map associated
                                  with this `Session`, and secondly by consulting the
                                  `Metadata` associated with the `Table` to which the
                                  `Mapper` is mapped for a bind.

        :param Union[ClauseElement, str] clause: a raw string query or a `ClauseElement`
                                                 (i.e. `~.sql.expression.select`,
                                                 `~.sql.expression.text`, etc.). if the
                                                 `Mapper` argument is not present or could not
                                                 produce a bind, the given expression construct
                                                 will be searched for a bound element, typically
                                                 a `Table` associated with `Metadata`.

        :rtype: Engine
        """

        if mapper is None and isinstance(clause, (TextClause, str)):
            tables = extractor_services.find_table_names(clause)
            if len(tables) > 0:
                return database_services.get_table_engine(tables[0])

        return super().get_bind(mapper, clause=clause)



# -*- coding: utf-8 -*-
"""
utils misc module.
"""

from pyrin.core.context import DTO
from pyrin.core.exceptions import CoreAttributeError


def set_attributes(instance, **kwargs):
    """
    sets the provided keyword arguments as attributes in given object instance.

    :param CoreObject instance: instance to set attributes for.

    :returns: instance with updated attributes.
    :rtype: CoreObject
    """

    if instance is None:
        return instance

    for name, value in kwargs.items():
        setattr(instance, name, value)

    return instance


def extract_attributes(instance, *attributes):
    """
    extracts all attributes with given names from provided object instance.

    :param CoreObject instance: instance to extract attributes from.
    :param attributes: list of all attribute names that should be extracted.

    :raises CoreAttributeError: core attribute error.

    :returns: dict(str name: object value)
    :rtype: dict
    """

    result = DTO()

    if instance is None:
        return result

    for name in attributes:
        if not hasattr(instance, name):
            raise CoreAttributeError('Object [{object}] has no attribute [{name}].'
                                     .format(object=str(instance), name=name))
        result[name] = getattr(instance, name)

    return result



# -*- coding: utf-8 -*-
"""
utils function module.
"""

import inspect


def get_doc(func, include_returns=True):
    """
    gets docstring of given function.

    :param function func: function to get its docstring.

    :param bool include_returns: specifies that `returns` and `rtype` parts
                                 of docstring must also be included in the result.
                                 defaults to True if not provided.

    :rtype: str
    """

    value = inspect.getdoc(func)
    if value is None:
        return None

    if include_returns is False:
        return_key = ':return'
        rtype_key = ':rtype'

        not_found_index = 5000000
        return_key_index = not_found_index
        rtype_key_index = not_found_index

        if return_key in value:
            return_key_index = value.index(return_key)

        if rtype_key in value:
            rtype_key_index = value.index(rtype_key)

        last_index = min(return_key_index, rtype_key_index)

        if last_index < not_found_index:
            value = value[0:last_index]

    return value



# -*- coding: utf-8 -*-
"""
model exceptions module.
"""

from pyrin.core.exceptions import CoreException


class DatabaseModelException(CoreException):
    """
    database model exception.
    """
    pass


class ColumnNotExistedError(DatabaseModelException):
    """
    column not existed error.
    """
    pass


class EntityNotHashableError(DatabaseModelException):
    """
    entity not hashable error.
    """
    pass



# -*- coding: utf-8 -*-
"""
database paging manager module.
"""

import pyrin.configuration.services as config_services

from pyrin.core.context import Manager


class DatabasePagingManager(Manager):
    """
    database paging manager class.
    """

    PAGING_KEY = '__paging__'
    LIMIT_KEY = '__limit__'
    OFFSET_KEY = '__offset__'

    def __init__(self):
        """
        initializes an instance of DatabasePagingManager.
        """

        super().__init__()

        self._limit = config_services.get('database', 'paging', 'limit')
        self._auto = config_services.get('database', 'paging', 'auto')

    def get_default_limit(self):
        """
        gets default limit from database config store.

        :rtype: int
        """

        return self._limit

    def is_auto_paging(self):
        """
        gets a value indicating that auto paging is enabled.

        :rtype: bool
        """

        return self._auto

    def enable_limit(self, options, limit=None):
        """
        enables limit with default or given limit parameter in given dict.

        :param int limit: custom limit value to be injected into given dict.
                          if not provided, defaults to limit value in
                          database config store.

        :param dict options: options dict to inject limit in it.
        """

        if limit is None:
            limit = self.get_default_limit()

        options[self.LIMIT_KEY] = limit

    def _enable_offset(self, options, offset):
        """
        enables offset with given value in input dict.

        :param int offset: offset value that should be injected into given dict.

        :param dict options: options dict to enable offset in it.
        """

        options[self.OFFSET_KEY] = offset

    def disable_limit(self, options):
        """
        disables limit in given dict. it removes limit key from it.

        :param dict options: options dict to disable limit in it.
        """

        options.pop(self.LIMIT_KEY, None)

    def _disable_offset(self, options):
        """
        disables offset in given dict. it removes offset key from it.

        :param dict options: options dict to disable offset in it.
        """

        options.pop(self.OFFSET_KEY, None)

    def extract_limit(self, options):
        """
        extracts and gets limit parameter from given dict.
        it returns None if no limit is set in given dict.

        :param dict options: options dict to get limit from it.

        :rtype: int
        """

        return options.get(self.LIMIT_KEY, None)

    def extract_offset(self, options):
        """
        extracts and gets offset parameter from given dict.
        it returns None if no offset is set in given dict.

        :param dict options: options dict to get offset from it.

        :rtype: int
        """

        return options.get(self.OFFSET_KEY, None)

    def enable_paging(self, options, offset, limit=None):
        """
        enables paging in given dict.
        it adds a paging key in given dict and sets its value to
        True. it also injects default limit and offset into it

        :param dict options: options dict to enable paging in it.
        :param int offset: offset value to be set in given dict.
        :param int limit: limit value to be set in given dict.
        """

        options[self.PAGING_KEY] = True
        self.enable_limit(options, limit)
        self._enable_offset(options, offset)

    def disable_paging(self, options):
        """
        disables paging in given dict.
        it removes paging key in given dict and also
        removes keys for limit and offset.

        :param dict options: options dict to disable paging in it.
        """

        options.pop(self.PAGING_KEY, None)
        self.disable_limit(options)
        self._disable_offset(options)

    def extract_paging(self, options):
        """
        extracts and gets limit and offset parameters from given dict.
        if limit or offset key is not available in provided
        dict, it gets the value from database config store
        from `paging` section for the absent key.

        :param dict options: options dict to get limit and offset from it.

        :returns: tuple(int limit, int offset)
        :rtype: tuple
        """

        limit = self.extract_limit(options)
        offset = self.extract_offset(options)

        return limit, offset

    def is_paging_required(self, options):
        """
        gets a value indicating that paging should be done.
        it checks that paging key name is available in dict
        and its value is set to True, otherwise returns False.

        :param dict options: options dict to detect paging should be done from it.

        :rtype: bool
        """

        paging = options.get(self.PAGING_KEY, False)
        return paging is True



# -*- coding: utf-8 -*-
"""
template package.
"""

from pyrin.packaging.base import Package


class TemplatePackage(Package):
    """
    template package class.
    """

    NAME = __name__
    COMPONENT_NAME = 'template.component'



# -*- coding: utf-8 -*-
"""
APPLICATION_PACKAGE package.
"""

from pyrin.application.base import Application


class APPLICATION_CLASS(Application):
    """
    server should create an instance of this class on startup.
    """
    pass
